/* PICTranslator.cpp */
/* PIC/PI=>BITMAP(32bit) */

#include	<Be.h>
#include	<TranslatorAddOn.h>

#define	VERSION	"ver.0.02     1999.4.26"
#define	BUFSIZE	0x10000

char	translatorName[]="PIC Translator";	
char	translatorInfo[]="PIC file format translator "VERSION;
int32	translatorVersion=000;

const char	rmode_name[2][16]={	"normal","correct aspect"	};
const char	wmode_name[1][16]={	"mac-15bit"	};

static char	rmode=1;
static char	wmode=0;

translation_format	inputFormats[]={
	{	B_TRANSLATOR_BITMAP,	B_TRANSLATOR_BITMAP,	0.5,0.5,	"image/x-be-bitmap",	"Be Bitmap Format"	},
	{	'PIC ',					B_TRANSLATOR_BITMAP,	0.5,0.5,	"image/pic",			"PIC bitmap format"	},
	{	'PI  ',					B_TRANSLATOR_BITMAP,	0.5,0.5,	"image/pi",				"Pi bitmap format"	},
	{0,0,0,0,{""},{""}}
};
translation_format	outputFormats[]={
	{	B_TRANSLATOR_BITMAP,	B_TRANSLATOR_BITMAP,	0.5,0.5,	"image/x-be-bitmap",	"Be Bitmap Format"	},
	{	'PIC ',					B_TRANSLATOR_BITMAP,	0.5,0.5,	"image/pic",			"PIC bitmap format"	},
//	{	'PI  ',					B_TRANSLATOR_BITMAP,	0.5,0.5,	"image/pi",				"Pi bitmap format"	},
	{0,0,0,0,{""},{""}}
};

struct PICChecker{		//PIC Checker format(size=4)
	char		mark[4];		//	"PIC","Pi"
};

struct PICHeader{		//PIC Header format(size=8)
	char		reserved;
	uchar		mode;			//	b7-4 mode  b3-0 hard
	uint16		depth;			//  0x0f
	uint16		width,height;
};

struct PIHeader{		//Pi Header format(size=14)
	uchar		mode;
	uchar		aspect_v,aspect_h;
	uchar		planes;
	char		hard[4];
	uint16		optsize;
	uint16		width,height;
};

struct cache{
	uchar	p,n;
	uint32	c;
};

uchar	*rb,*wb;				//buffer
uint	rp,wp;
uchar	rpb,wpb;
int		rbl,wbl;
off_t	sss;			//write size
//-------------------------------------------------------------------------------------
class PICWindow:public BWindow
{
	public:
		PICWindow(BRect area,const char *title,window_type kind,uint32 flags);
		bool	QuitRequested();
};

class configview:public BView
{
	public:
		configview(BRect frame);
		void	AttachedToWindow();
		void	MessageReceived(BMessage *msg);
	private:
		BRadioButton	*rb1[2],*rb2[2];
};

PICWindow::PICWindow(BRect area,const char *title,window_type kind,uint32 flags):BWindow(area,title,kind,flags)
{
}

bool PICWindow::QuitRequested()
{
	be_app->PostMessage(B_QUIT_REQUESTED);
	return(true);
}

configview::configview(BRect frame):BView(frame,"PIC_Translator Settings",B_NOT_RESIZABLE,B_WILL_DRAW)
{
	BStringView	*sv;
	BMessage	*mes;
	BBox		*box;
	sv=new BStringView(BRect(10,3,190,20),"title","PICTranslator  (PIC/Pi)");
	sv->SetFont(be_bold_font);
	AddChild(sv);
	sv=new BStringView(BRect(40,20,180,37),"ver",VERSION);
	sv->SetFont(be_bold_font);
	AddChild(sv);
	sv=new BStringView(BRect(10,113,190,130),"program","code:  (c)Hiroyuki Tsutsumi");
	sv->SetFont(be_bold_font);
	AddChild(sv);
	sv=new BStringView(BRect(10,130,190,147),"format","format:  (c)Yanagisawa");
	sv->SetFont(be_bold_font);
	AddChild(sv);
	box=new BBox(BRect(0,48,100,102),"box1");
	box->SetLabel("read(PIC)");
	for(int i=0;i<2;i++){
		mes=new BMessage('rmod');mes->AddInt8("aspect_mode",i);
		rb1[i]=new BRadioButton(BRect(5,i*15+15,95,i*15+30),"radiobutton",rmode_name[i],mes);
		box->AddChild(rb1[i]);
	}
	AddChild(box);
	box=new BBox(BRect(100,48,200,102),"box2");
	box->SetLabel("write(PIC)");
	for(int i=0;i<1;i++){
		mes=new BMessage('wmod');mes->AddInt8("encode_mode",i);
		rb2[i]=new BRadioButton(BRect(5,i*15+15,95,i*15+30),"radiobutton",wmode_name[i],mes);
		box->AddChild(rb2[i]);
	}
	AddChild(box);
}

void configview::AttachedToWindow()
{
	SetViewColor(233,255,233);
	SetLowColor(122,155,122);
	for(int i=0;i<2;i++){
		rb1[i]->SetTarget(this);
		if(rmode==i)rb1[i]->SetValue(1);
	}
	for(int i=0;i<1;i++){
		rb2[i]->SetTarget(this);
		if(wmode==i)rb2[i]->SetValue(1);
	}
}

void configview::MessageReceived(BMessage *msg)
{
	switch(msg->what){
		case 'rmod':{
			beep();
			msg->FindInt8("aspect_mode",(signed char*)&rmode);
		}break;
		case 'wmod':{
			beep();
			msg->FindInt8("encode_mode",(signed char*)&wmode);
		}break;
		default:	BView::MessageReceived(msg);
	}
}
//-------------------------------------------------------------------------------------
main()
{
	BApplication app("application/x-vnd.Beer-PIC_Translator");
	PICWindow *window=new PICWindow(BRect(100,100,400,300),"PIC Settings",B_TITLED_WINDOW,B_NOT_RESIZABLE|B_NOT_ZOOMABLE);
	BView *config=NULL;
	BRect extent;
	MakeConfig(NULL,&config,&extent);
	window->ResizeTo(extent.Width(),extent.Height());
	window->AddChild(config);
	window->Show();
	app.Run();
}

status_t Identify(BPositionIO *inSource,const translation_format *inFormat,BMessage *ioExtension,translator_info *outInfo,uint32 outType)
{
	if((outType)&&(outType!=B_TRANSLATOR_BITMAP)&&(outType!='PIC '))return(B_NO_TRANSLATOR);
	const int			bufsize=32;
	char				buf[bufsize];
	TranslatorBitmap	*bmphd=(TranslatorBitmap*)buf;
	PICChecker			*picck=(PICChecker*)buf;
	PICHeader			*pichd=(PICHeader*)(buf+4);
	PIHeader			*pihd=(PIHeader*)(buf+4);

	//	read checker(PIC/Pi)/header(TranslatorBitmap)
	ssize_t		len,stt;
	memset(buf,0,4);
	stt=inSource->Read(buf,len=4);
	if(stt<4)return(stt);

	//	check input-type
	int			itype=-1;
	if(B_BENDIAN_TO_HOST_INT32(bmphd->magic)==B_TRANSLATOR_BITMAP){
		itype=0;						//B_TRANSLATOR_BITMAP
	}else if(!strncmp(picck->mark,"PIC",3)){
		itype=1;						//PIC
	}else if(!strncmp(picck->mark,"Pi",2)){
		itype=2;						//Pi
	}else return(B_NO_TRANSLATOR);

	//	read header
	if(itype==0){						//B_TRANSLATOR_BITMAP
		len=sizeof(TranslatorBitmap)-4;
		stt=inSource->Read((char*)bmphd+4,len);
		if(stt<len)return(stt);
	}else if(itype==1){					//PIC
		while(1){
			stt=inSource->Read(pichd,1);
			if(stt<1)return(stt);
			if(pichd->reserved==0x00)break;
		}
		len=sizeof(PICHeader);
		stt=inSource->Read(pichd,len);
		if(stt<len)return(stt);
	}else if(itype==2){					//Pi
		if(picck->mark[3])while(1){
			stt=inSource->Read(pihd,1);
			if(stt<1)return(stt);
			if(pihd->mode==0x00)break;
		}
		len=sizeof(PIHeader);
		stt=inSource->Read(pihd,len);
		if(stt<len)return(stt);
	}

	//	check header & out info
	if(itype==0){						//B_TRANSLATOR_BITMAP
		if(B_BENDIAN_TO_HOST_INT32(bmphd->colors)!=B_RGB_32_BIT)return(B_NO_TRANSLATOR);
		sprintf(outInfo->name,"%s, type %d",inputFormats[0].name,B_BENDIAN_TO_HOST_INT32(bmphd->colors));
	}else if(itype==1){					//PIC
		if((B_BENDIAN_TO_HOST_INT16(pichd->depth)!=15)&&(B_BENDIAN_TO_HOST_INT16(pichd->depth)!=16))return(B_NO_TRANSLATOR);
		if(((pichd->mode&0x0f)!=0)&&((pichd->mode&0x0f)!=2)&&((pichd->mode&0x0f)!=3))return(B_NO_TRANSLATOR);
		sprintf(outInfo->name,"%s, mode %02x(%dbits)",inputFormats[1].name,pichd->mode,B_BENDIAN_TO_HOST_INT16(pichd->depth));
	}else if(itype==2){					//Pi
		if(B_BENDIAN_TO_HOST_INT16(pihd->optsize)!=0)return(B_NO_TRANSLATOR);
		sprintf(outInfo->name,"%s, (%dcolors)",inputFormats[2].name,1<<pihd->planes);
	}
	outInfo->type=inputFormats[itype].type;
	outInfo->translator=0;
	outInfo->group=inputFormats[itype].group;
	outInfo->quality=inputFormats[itype].quality;
	outInfo->capability=inputFormats[itype].capability;
	strcpy(outInfo->MIME,inputFormats[itype].MIME);
	return(B_OK);
}

static status_t CopyLoop(BPositionIO *input,BPositionIO *output)
{																//BITMAP->BITMAP , PIC->PIC
	rb=new uchar[BUFSIZE];			if(!rb)return(B_NO_MEMORY);
	ssize_t		len,stt;
	while((stt=input->Read(rb,len=BUFSIZE))>0)if((stt=output->Write(rb,len=stt))<0)break;
	return(stt);
}

static status_t writebits(BPositionIO *output,int c,uint32 d)
{
	status_t	len,stt;
	if(c==0){
		d++;
		uint32	dd=d;
		c=0;
		while(dd>>=1)c++;
		for(int i=c;i>0;i--){
			if(i>1)wb[wp]|=wpb;
			if(!(wpb>>=1)){
				wpb=0x80;
				if(++wp==BUFSIZE){
					stt=output->Write(wb,len=wp);
					if(stt<len)return(stt);
					sss+=wp;wp=0;
					memset(wb,0,BUFSIZE);
				}
			}
		}
	}
	if(c>0){
		uint32	bb=1<<c;
		while(bb>>=1){
			if(d&bb)wb[wp]|=wpb;
			if(!(wpb>>=1)){
				wpb=0x80;
				if(++wp==BUFSIZE){
					stt=output->Write(wb,len=wp);
					if(stt<len)return(stt);
					sss+=wp;wp=0;
					memset(wb,0,BUFSIZE);
				}
			}
		}
	}
	if(c<0){		//	flush & setsize
		if(wpb<0x80){wp++;wpb=0x80;}
		if(wp>0){
			stt=output->Write(wb,len=wp);
			if(stt<len)return(stt);
			sss+=wp;wp=0;
		}
		stt=output->SetSize(sss);
		if(stt<B_OK)return(stt);
	}
	return(B_OK);
}

static status_t WritePIC(BPositionIO *input,BPositionIO *output,BMessage *ioExtension)
{																//BITMAP->PIC
	TranslatorBitmap	bmphd;
	PICChecker			picck;
	PICHeader			pichd;
	uint32				pal[256];
	char		mode;
	ssize_t		len,stt;
	int			w,h;

	//	read bitmap header
	stt=input->Read(&bmphd,len=sizeof(bmphd));
	if(stt<len)return(stt);
	bmphd.magic=B_BENDIAN_TO_HOST_INT32(bmphd.magic);
	bmphd.bounds.left=B_BENDIAN_TO_HOST_FLOAT(bmphd.bounds.left);
	bmphd.bounds.top=B_BENDIAN_TO_HOST_FLOAT(bmphd.bounds.top);
	bmphd.bounds.right=B_BENDIAN_TO_HOST_FLOAT(bmphd.bounds.right);
	bmphd.bounds.bottom=B_BENDIAN_TO_HOST_FLOAT(bmphd.bounds.bottom);
//	bmphd.colors=B_BENDIAN_TO_HOST_INT32(bmphd.colors);
	*(int32*)(&(bmphd.colors))=B_BENDIAN_TO_HOST_INT32(bmphd.colors);
	bmphd.rowBytes=B_BENDIAN_TO_HOST_INT32(bmphd.rowBytes);
	bmphd.dataSize=B_BENDIAN_TO_HOST_INT32(bmphd.dataSize);
	if(bmphd.colors!=B_RGB_32_BIT)return(B_NO_TRANSLATOR);
	//	write PIC header
	strncpy(picck.mark,"PIC",3);
	picck.mark[3]=0x1a;
	stt=output->Write(&picck,len=sizeof(picck));
	if(stt<len)return(stt);
	pichd.reserved=0x00;
	stt=output->Write(&pichd,len=1);
	if(stt<len)return(stt);
	pichd.mode=0x03;							//mac-mode
	pichd.depth=B_HOST_TO_BENDIAN_INT16(15);	//15bit-color
	pichd.width=B_HOST_TO_BENDIAN_INT16(w=bmphd.bounds.right-bmphd.bounds.left+1);
	pichd.height=B_HOST_TO_BENDIAN_INT16(h=bmphd.bounds.bottom-bmphd.bounds.top+1);
	stt=output->Write(&pichd,len=sizeof(pichd));
	if(stt<len)return(stt);
	sss=sizeof(picck)+1+sizeof(pichd);

	//	xfer data
	rb=new uchar[bmphd.dataSize];	if(!rb)return(B_NO_MEMORY);
	wb=new uchar[BUFSIZE];			if(!wb)return(B_NO_MEMORY);
	stt=input->Read(rb,len=bmphd.dataSize);		//read all
	if(stt<len)return(stt);
	memset(wb,0,BUFSIZE);
	wp=0;wpb=0x80;
	uint32	*rbp=(uint32*)rb;
	for(int z=0;z<w*h;z++){
		uint32	c=B_BENDIAN_TO_HOST_INT32(*rbp);
		uchar	b=(c>>27)&0x1f;
		uchar	g=(c>>19)&0x1f;
		uchar	r=(c>>11)&0x1f;
		*(rbp++)=((uint32)r<<10)|((uint32)g<<5)|((uint32)b);
	}
	uint32	c;
	cache	*ct=new cache[128];	if(!ct)return(B_NO_MEMORY);
	for(int i=0;i<128;i++){
		ct[i].p=(i+1)&127;
		ct[i].n=(i+127)&127;
		ct[i].c=0xffffffff;
	}
	uchar	last=0;
	rbp=(uint32*)rb;
	c=*rbp;	(*rbp)|=0x80000000;	rbp++;
	for(int z=1;z<w*h;z++){
		if(c!=*rbp){
			c=*rbp;	(*rbp)|=0x80000000;
		}
		rbp++;
	}
	rbp=(uint32*)rb;
	int	l=1;
	for(int y=0;y<h;y++)for(int x=0;x<w;x++){
		c=*rbp;
		if(c&0x80000000){
			writebits(output,0,l);							//length
			l=1;
			int	i;
			c&=0x7fffffff;
			for(i=0;i<128;i++)if(ct[i].c==c)break;
			if(i<128){		//hit cash
				writebits(output,1,1);					//1
				writebits(output,7,i);						//cash_no
				if(i!=last){
					ct[ct[i].p].n=ct[i].n;
					ct[ct[i].n].p=ct[i].p;
					ct[ct[last].p].n=i;
					ct[i].p=ct[last].p;
					ct[i].n=last;
					ct[last].p=i;
					last=i;
				}
			}else{			//new
				writebits(output,1,0);					//0
				writebits(output,15,c);						//color
				last=ct[last].p;
				ct[last].c=c;
			}
			c|=0x80000000;
			uint32	*rbp0=rbp;
			for(int	yy=y+1;;yy++){
				uint32	*rbp1;
				int	d=0;
				if(yy<h){
					rbp0+=w;
					if(*(rbp1=rbp0)==c)d=1;
					else if((x>0)&&(*(rbp1=rbp0-1)==c))d=2;
					else if((x<w-1)&&(*(rbp1=rbp0+1)==c))d=3;
					else if((x>1)&&(*(rbp1=rbp0-2)==c))d=4;
					else if((x<w-2)&&(*(rbp1=rbp0+2)==c))d=5;
				}
				if(d==0){
					if(yy==y+1)writebits(output,1,0);	//0
					else writebits(output,3,0);				//000
					break;
				}else{
					if(yy==y+1)writebits(output,1,1);	//1
					if(d==1)writebits(output,2,2);			//10
					else if(d==2)writebits(output,2,1);		//01
					else if(d==3)writebits(output,2,3);		//11
					else if(d==4)writebits(output,4,2);		//0010
					else writebits(output,4,3);				//0011
					*(rbp0=rbp1)&=0x7fffffff;
				}
			}
		}else{
			l++;
		}
		rbp++;
	}
	writebits(output,0,l);									//length
	delete	ct;
	writebits(output,-1,0);
	return(B_OK);
}

static status_t readbits(BPositionIO *input,int c,uint32 *d)
{
	status_t	len,stt;
	uint32		dd=0;
	*d=0;
	if(c<=0){
		int	cc=1;
		if(c==0){
			dd=-1;
		}else if(c<-1){
			if(rpb==1){
				rpb=0x80;
				rp++;
				if(rp==rbl){
					stt=input->Read(rb,len=BUFSIZE);
					if(stt<0)return(stt);
					rbl=stt;rp=0;
				}
			}else{
				rpb>>=1;
			}
			if(rb[rp]&rpb){dd=-2;cc=0;c=1;}else{cc=c+2;c=0;}
		}
		if(cc)while(1){
			if(rpb==1){
				rpb=0x80;
				rp++;
				if(rp==rbl){
					stt=input->Read(rb,len=BUFSIZE);
					if(stt<0)return(stt);
					rbl=stt;rp=0;
				}
			}else{
				rpb>>=1;
			}
			c++;
			if((rb[rp]&rpb)==0)break;
			cc++;
			if(cc==0){c++;break;}
		}
		dd+=1<<c;
	}
	uint32	bb=1<<(c-1);
	while(c-->0){
		if(rpb==1){
			rpb=0x80;
			rp++;
			if(rp==rbl){
				stt=input->Read(rb,len=BUFSIZE);
				if(stt<0)return(stt);
				rbl=stt;rp=0;
			}
		}else{
			rpb>>=1;
		}
		if(rb[rp]&rpb)*d|=bb;
		bb>>=1;
	}
	*d+=dd;
	return(B_OK);
}

static status_t ReadPIC(BPositionIO *input,BPositionIO *output,BMessage *ioExtension)
{																//PIC->BITMAP
	TranslatorBitmap	bmphd;
	PICChecker			picck;
	uint32				pal[256];
	char		mode;
	ssize_t		len,stt;
	int			w,h;

	//	read PIC checker
	stt=input->Read(&picck,len=sizeof(picck));
	if(stt<len)return(stt);
	if(!strncmp(picck.mark,"PIC",3))mode=1;
	else if(!strncmp(picck.mark,"Pi",2))mode=2;
	else return(B_NO_TRANSLATOR);

	if(mode==1){						//PIC
		//  read PIC header
		PICHeader			pichd;
		while(1){
			stt=input->Read(&pichd,1);
			if(stt<1)return(stt);
			if(pichd.reserved==0x00)break;
		}
		stt=input->Read(&pichd,len=sizeof(pichd));
		if(stt<len)return(stt);
		pichd.depth=B_BENDIAN_TO_HOST_INT16(pichd.depth);
		w=pichd.width=B_BENDIAN_TO_HOST_INT16(pichd.width);
		h=pichd.height=B_BENDIAN_TO_HOST_INT16(pichd.height);
		if(((pichd.mode&0x0f)!=0)&&((pichd.mode&0x0f)!=2)&&((pichd.mode&0x0f)!=3))return(B_NO_TRANSLATOR);
		if((pichd.depth!=15)&&(pichd.depth!=16))return(B_NO_TRANSLATOR);

		if((rmode>0)&&(pichd.mode==0x00))w=w*4/3;		//for aspect(4:3)

		//	write bitmap header(output=32bit only)
		bmphd.rowBytes=w*4;
		bmphd.dataSize=bmphd.rowBytes*h;
		stt=output->SetSize(sizeof(bmphd)+bmphd.dataSize);
		if(stt<B_OK)return(stt);
		bmphd.magic=B_HOST_TO_BENDIAN_INT32(B_TRANSLATOR_BITMAP);
		bmphd.bounds.left=B_HOST_TO_BENDIAN_FLOAT((float)0);
		bmphd.bounds.top=B_HOST_TO_BENDIAN_FLOAT((float)0);
		bmphd.bounds.right=B_HOST_TO_BENDIAN_FLOAT((float)w-1);
		bmphd.bounds.bottom=B_HOST_TO_BENDIAN_FLOAT((float)h-1);
//		bmphd.colors=B_HOST_TO_BENDIAN_INT32(B_RGB_32_BIT);
		*(int32*)(&(bmphd.colors))=B_HOST_TO_BENDIAN_INT32(B_RGB_32_BIT);
		bmphd.rowBytes=B_HOST_TO_BENDIAN_INT32(bmphd.rowBytes);
		bmphd.dataSize=B_HOST_TO_BENDIAN_INT32(bmphd.dataSize);
		stt=output->Write(&bmphd,len=sizeof(bmphd));
		if(stt<len)return(stt);
		bmphd.magic=B_BENDIAN_TO_HOST_INT32(bmphd.magic);
		bmphd.bounds.left=B_BENDIAN_TO_HOST_FLOAT(bmphd.bounds.left);
		bmphd.bounds.top=B_BENDIAN_TO_HOST_FLOAT(bmphd.bounds.top);
		bmphd.bounds.right=B_BENDIAN_TO_HOST_FLOAT(bmphd.bounds.right);
		bmphd.bounds.bottom=B_BENDIAN_TO_HOST_FLOAT(bmphd.bounds.bottom);
//		bmphd.colors=B_BENDIAN_TO_HOST_INT32(bmphd.colors);
		*(int32*)(&(bmphd.colors))=B_BENDIAN_TO_HOST_INT32(bmphd.colors);
		bmphd.rowBytes=B_BENDIAN_TO_HOST_INT32(bmphd.rowBytes);
		bmphd.dataSize=B_BENDIAN_TO_HOST_INT32(bmphd.dataSize);

		//	xfer data
		rb=new uchar[BUFSIZE];			if(!rb)return(B_NO_MEMORY);
		wb=new uchar[bmphd.dataSize];	if(!wb)return(B_NO_MEMORY);
		memset(wb,0,bmphd.dataSize);	//clr output-area
		rbl=0;rp=-1;rpb=1;
		w=pichd.width;	h=pichd.height;
		cache	*ct=new cache[128];	if(!ct)return(B_NO_MEMORY);
		for(int i=0;i<128;i++){
			ct[i].p=(i+1)&127;
			ct[i].n=(i+127)&127;
			ct[i].c=0;
		}
		uchar	last=0;
		uint32	*wbp=(uint32*)wb;
		uint32	r,c=0;
		int		l=1;
//		for(int y=0;y<h;y++)for(int x=0;x<w;x++){
		for(int z=0;z<w*h;z++){
			if(--l==0){
				readbits(input,0,&r);
				l=r;
			}
			if(l>1){
				if(*wbp)c=(*wbp)&0x7fffffff;
				*wbp=c;
			}else{
				readbits(input,1,&r);
				if(r){
					readbits(input,7,&r);
					if(r!=last){
						ct[ct[r].p].n=ct[r].n;
						ct[ct[r].n].p=ct[r].p;
						ct[ct[last].p].n=r;
						ct[r].p=ct[last].p;
						ct[r].n=last;
						ct[last].p=r;
						last=r;
					}
					c=ct[r].c;
				}else{
					readbits(input,pichd.depth,&r);
					c=r;
					last=ct[last].p;
					ct[last].c=c;
				}
				*wbp=c;
				readbits(input,1,&r);
				if(r){
					uint32	*wbp0=wbp;
					while(1){
						readbits(input,2,&r);
						if(r==0){
							readbits(input,1,&r);
							if(r==0)break;
							readbits(input,1,&r);
							wbp0+=r?2:-2;
						}else if(r==1){
							wbp0--;
						}else if(r==3){
							wbp0++;
						}
						wbp0+=w;
						*wbp0=c|0x80000000;
					}
				}
			}
			wbp++;
		}
		delete	ct;
		wbp=(uint32*)wb;
		for(int z=0;z<w*h;z++){
			c=*wbp;
			uchar	c1,c2,c3;
			if(pichd.depth==16){			//16bit color
				c1=(c>>10)&0x3e|(c&1);	c1=(c1<<2)|(c1>>4);
				c2=(c>>5)&0x3e|(c&1);	c2=(c2<<2)|(c2>>4);
				c3=c&0x3f;				c3=(c3<<2)|(c3>>4);
			}else{							//15bit color
				c1=(c>>10)&0x1f;		c1=(c1<<3)|(c1>>2);
				c2=(c>>5)&0x1f;			c2=(c2<<3)|(c2>>2);
				c3=c&0x1f;				c3=(c3<<3)|(c3>>2);
			}
			if((pichd.mode&0x0f)==3){		//mac
				c=(((uint32)c3)<<24)|(((uint32)c2)<<16)|(((uint32)c1)<<8);
			}else{							//x68k,towns
				c=(((uint32)c3)<<24)|(((uint32)c1)<<16)|(((uint32)c2)<<8);
			}
			*(wbp++)=B_HOST_TO_BENDIAN_INT32(c);
		}
		if((rmode>0)&&(pichd.mode==0x00)){			//for aspect(4:3)
			wbp=(uint32*)wb;
			for(int	y=0;y<h;y++){
				uint32	*rbp=(uint32*)rb;
				uchar	i=0;
				for(int x=0;x<w;x++){
					if(i++==2){*(rbp++)=*wbp;i=0;}
					*(rbp++)=*(wbp++);
				}
				stt=output->Write(rb,len=bmphd.rowBytes);
				if(stt<len)return(stt);
			}
		}else{
			stt=output->Write(wb,len=bmphd.dataSize);
			if(stt<len)return(stt);
		}
	}else{								//Pi
		//  read Pi header
		PIHeader			pihd;
		if(picck.mark[3])while(1){
			stt=input->Read(&pihd,1);
			if(stt<1)return(stt);
			if(pihd.mode==0x00)break;
		}
		stt=input->Read(&pihd,len=sizeof(pihd));
		if(stt<len)return(stt);
		pihd.optsize=B_BENDIAN_TO_HOST_INT16(pihd.optsize);
		if(pihd.optsize!=0)return(B_NO_TRANSLATOR);
		pihd.width=B_BENDIAN_TO_HOST_INT16(pihd.width);
		pihd.height=B_BENDIAN_TO_HOST_INT16(pihd.height);
		//	write bitmap header(output=32bit only)
		bmphd.rowBytes=pihd.width*4;
		bmphd.dataSize=bmphd.rowBytes*pihd.height;
		stt=output->SetSize(sizeof(bmphd)+bmphd.dataSize);
		if(stt<B_OK)return(stt);
		bmphd.magic=B_HOST_TO_BENDIAN_INT32(B_TRANSLATOR_BITMAP);
		bmphd.bounds.left=B_HOST_TO_BENDIAN_FLOAT((float)0);
		bmphd.bounds.top=B_HOST_TO_BENDIAN_FLOAT((float)0);
		bmphd.bounds.right=B_HOST_TO_BENDIAN_FLOAT((float)pihd.width-1);
		bmphd.bounds.bottom=B_HOST_TO_BENDIAN_FLOAT((float)pihd.height-1);
//		bmphd.colors=B_HOST_TO_BENDIAN_INT32(B_RGB_32_BIT);
		*(int32*)(&(bmphd.colors))=B_HOST_TO_BENDIAN_INT32(B_RGB_32_BIT);
		bmphd.rowBytes=B_HOST_TO_BENDIAN_INT32(bmphd.rowBytes);
		bmphd.dataSize=B_HOST_TO_BENDIAN_INT32(bmphd.dataSize);
		stt=output->Write(&bmphd,len=sizeof(bmphd));
		if(stt<len)return(stt);
		bmphd.magic=B_BENDIAN_TO_HOST_INT32(bmphd.magic);
		bmphd.bounds.left=B_BENDIAN_TO_HOST_FLOAT(bmphd.bounds.left);
		bmphd.bounds.top=B_BENDIAN_TO_HOST_FLOAT(bmphd.bounds.top);
		bmphd.bounds.right=B_BENDIAN_TO_HOST_FLOAT(bmphd.bounds.right);
		bmphd.bounds.bottom=B_BENDIAN_TO_HOST_FLOAT(bmphd.bounds.bottom);
//		bmphd.colors=B_BENDIAN_TO_HOST_INT32(bmphd.colors);
		*(int32*)(&(bmphd.colors))=B_BENDIAN_TO_HOST_INT32(bmphd.colors);
		bmphd.rowBytes=B_BENDIAN_TO_HOST_INT32(bmphd.rowBytes);
		bmphd.dataSize=B_BENDIAN_TO_HOST_INT32(bmphd.dataSize);

		//	xfer data
		rb=new uchar[BUFSIZE];			if(!rb)return(B_NO_MEMORY);
		wb=new uchar[bmphd.dataSize];	if(!wb)return(B_NO_MEMORY);
		memset(wb,0,bmphd.dataSize);	//clr output-area
		rbl=0;rp=-1;rpb=1;
		w=pihd.width;	h=pihd.height;
		int	cols=1<<pihd.planes;
		if(pihd.mode&0x80){						//  default palette
			for(int i=0;i<cols;i++){
				uchar	r,g,b;
				if(cols==16){
					r=g=b=0;
					if(i&2)r=(i<8)?0x7f:0xff;
					if(i&4)g=(i<8)?0x7f:0xff;
					if(i&1)b=(i<8)?0x7f:0xff;
				}else{
					r=(i>>2)&0x07;	r=(r<<5)|(r<<2)|(r>>1);
					g=(i>>5)&0x07;	g=(g<<5)|(g<<2)|(g>>1);
					b=i&0x03;		b=(b<<6)|(b<<4)|(b<<2)|b;
				}
				pal[i]=B_HOST_TO_BENDIAN_INT32((((uint32)b)<<24)|(((uint32)g)<<16)|(((uint32)r)<<8));
			}
		}else{									//  read palette
			stt=input->Read(pal,len=3*cols);
			if(stt<len)return(stt);
			for(int i=cols;i>=0;i--){
				uchar	r=*((uchar*)pal+i*3);
				uchar	g=*((uchar*)pal+i*3+1);
				uchar	b=*((uchar*)pal+i*3+2);
				pal[i]=B_HOST_TO_BENDIAN_INT32((((uint32)b)<<24)|(((uint32)g)<<16)|(((uint32)r)<<8));
			}
		}
		uchar	*ct=new uchar[cols*cols];	if(!ct)return(B_NO_MEMORY);
		for(int i=0;i<cols;i++)for(int j=0;j<cols;j++)ct[i*cols+j]=(cols+i-j)%cols;
		uint32	*wbp=(uint32*)wb;
		uint32	r;
		uchar	b,c,d;
		b=0;d=-1;
		readbits(input,-pihd.planes,&r);
		c=ct[b*cols+r];
		for(int i=r;i>0;i--)ct[b*cols+i]=ct[b*cols+i-1];ct[b*cols]=c;
		*wbp=c;
		b=c;
		readbits(input,-pihd.planes,&r);
		c=ct[b*cols+r];
		for(int i=r;i>0;i--)ct[b*cols+i]=ct[b*cols+i-1];ct[b*cols]=c;
		*(wbp+1)=c;
		b=c;
		for(int z=0;z<w*h;){
			readbits(input,2,&r);			//houkou
			if(r==3){readbits(input,1,&r);r+=3;}
			if(r==d){
				while(1){
					readbits(input,-pihd.planes,&r);
					c=ct[b*cols+r];
					for(int i=r;i>0;i--)ct[b*cols+i]=ct[b*cols+i-1];ct[b*cols]=c;
					*(wbp++)=c;
					b=c;z++;
					readbits(input,-pihd.planes,&r);
					c=ct[b*cols+r];
					for(int i=r;i>0;i--)ct[b*cols+i]=ct[b*cols+i-1];ct[b*cols]=c;
					*(wbp++)=c;
					b=c;z++;
					readbits(input,1,&r);
					if(r==0)break;
				}
				d=-1;
			}else{
				d=r;
				int	zz;
				if(d==0)zz=((*(wbp-2)==*(wbp-1))?-2:-4);
				else if(d==1)zz=-w;
				else if(d==2)zz=-w*2;
				else if(d==3)zz=-w+1;
				else if(d==4)zz=-w-1;
				readbits(input,-1,&r);		//nagasa
				r*=2;
				for(int j=0;j<r;j++){
					c=(wbp+zz>=(uint32*)wb)?*(wbp+zz):*((uint32*)wb+(((uint32*)wb-(wbp+zz))&1));
					*(wbp++)=c;
				}
				b=c;
				z+=r;
			}
		}
		for(int z=0;z<w*h;z++){
			c=*((uint32*)wb+z);
			*((uint32*)wb+z)=pal[c];
		}
		delete	ct;
		stt=output->Write(wb,len=bmphd.dataSize);
		if(stt<len)return(stt);
	}
	return(B_OK);
}

status_t Translate(BPositionIO *inSource,const translator_info *inInfo,BMessage *ioExtension,uint32 outType,BPositionIO *outDestination)
{
	status_t	stt=B_NO_TRANSLATOR;
	rb=wb=NULL;
	if(outType==inInfo->type)stt=CopyLoop(inSource,outDestination);
	if((outType=='PIC ')&&(inInfo->type==B_TRANSLATOR_BITMAP))stt=WritePIC(inSource,outDestination,ioExtension);
	if((outType==B_TRANSLATOR_BITMAP)&&(inInfo->type=='PIC '))stt=ReadPIC(inSource,outDestination,ioExtension);
	if((outType==B_TRANSLATOR_BITMAP)&&(inInfo->type=='PI  '))stt=ReadPIC(inSource,outDestination,ioExtension);
	delete rb;delete wb;
	return(stt);
}

status_t MakeConfig(BMessage *ioExtension,BView **outView,BRect *outExtent)
{
	*outView=new configview(BRect(0,0,200,150));
	outExtent->Set(0,0,200,150);
	return(B_OK);
}

status_t GetConfigMessage(BMessage *ioExtension)
{
	return(B_OK);
}
